from dataclasses import dataclass
from typing import Optional

from asyncpg import Connection  # type: ignore[import-untyped]


@dataclass
class Document:
    key: str
    token_count: int


@dataclass
class SearchResult:
    selected_documents: list[Document]
    overflowed_document: Optional[Document]


async def delete_document(connection: Connection, document_key: str) -> bool:
    '''
    Deletes the rows referecing the given document from all database tables, returning true if all records
    were found when deleting.
    '''

    status1 = await connection.execute(
        'DELETE FROM sbretrieval_embeddings WHERE document_key = $1', document_key
    )

    status2 = await connection.execute(
        'DELETE FROM sbretrieval_summaries WHERE document_key = $1', document_key
    )

    status3 = await connection.execute(
        'DELETE FROM sbretrieval_documents WHERE document_key = $1', document_key
    )

    # unfortunately asyncpg doesn't return the affected rows, only the status line string
    # see https://github.com/MagicStack/asyncpg/issues/311
    def affected_rows(s: str) -> int:
        tokens = s.strip().split()
        last = tokens[-1]
        return int(last)

    affected1 = affected_rows(status1)
    affected2 = affected_rows(status2)
    affected3 = affected_rows(status3)

    # returning true only if all three tables had their rows deleted
    return all([affected1, affected2, affected3])


async def search_documents_by_embeddings(
    connection: Connection,
    embedding: list[float],
    max_token_count: int
) -> SearchResult:
    '''Performs a search by embedding proximity.'''

    sql = '''
        SELECT d.document_key, d.document_token_count
        FROM sbretrieval_documents d
        JOIN sbretrieval_embeddings e ON d.document_key = e.document_key
        ORDER BY e.embedding <=> $1::vector -- "<=>" for cosine distance
    '''

    # formatting embedding as a string expected by pgvector
    embedding_string = vector_to_string(embedding)

    # sending the query and opening a cursor at the server, prefetching one row at a time
    statement = await connection.prepare(sql)
    cursor = statement.cursor(embedding_string, prefetch=1)

    # those variables will be filled in the loop bellow and returned at the end of the function
    document_keys = []
    overflowed_document = None
    total_token_count = 0

    # at each iteration asks another row from the open cursor at the server
    async for row in cursor:

        # formatting the document as the named tuple expected to be returned
        current_document = Document(row['document_key'], row['document_token_count'])

        # checking if, on this iteration, the total count overflows the maximum passed
        token_count_overflow = (total_token_count + current_document.token_count) > max_token_count

        # if yes, assigning the overflowed document to a different variable instead of adding to the list
        # and exiting the loop
        if token_count_overflow:
            overflowed_document = current_document
            break

        # if we don't have an overflow, just adds the current document and token count and go forward to
        # the next iteration
        document_keys.append(current_document)
        total_token_count += current_document.token_count

    # once outside the loop no further work is need, returned the values as the expected named tuple
    return SearchResult(document_keys, overflowed_document)


async def upsert_document(connection: Connection, document_key: str, document_token_count: int) -> None:
    '''Inserts into documents table.'''

    sql = '''
        INSERT INTO sbretrieval_documents (document_key, document_token_count)
        VALUES ($1, $2)
        ON CONFLICT (document_key)
        DO UPDATE SET document_token_count = EXCLUDED.document_token_count
    '''

    await connection.execute(sql, document_key, document_token_count)


async def upsert_document_embedding(
    connection: Connection,
    document_key: str,
    embedding: list[float],
    ai_vendor: str,
    model: str
) -> None:
    '''Inserts into embeddings table. '''

    sql = '''
        INSERT INTO sbretrieval_embeddings (document_key, embedding, ai_vendor, model)
        VALUES ($1, $2::vector, $3, $4)
        ON CONFLICT (document_key)
        DO UPDATE SET
            embedding = EXCLUDED.embedding,
            ai_vendor = EXCLUDED.ai_vendor,
            model = EXCLUDED.model
    '''

    embedding_string = vector_to_string(embedding)

    await connection.execute(sql, document_key, embedding_string, ai_vendor, model)


async def upsert_document_summary(
    connection: Connection,
    document_key: str,
    summary: str,
    ai_vendor: str,
    model: str
) -> None:
    '''Inserts into summaries table.'''

    sql = '''
        INSERT INTO sbretrieval_summaries (document_key, summary, ai_vendor, model)
        VALUES ($1, $2, $3, $4)
        ON CONFLICT (document_key)
        DO UPDATE SET
            summary = EXCLUDED.summary,
            ai_vendor = EXCLUDED.ai_vendor,
            model = EXCLUDED.model
    '''

    await connection.execute(sql, document_key, summary, ai_vendor, model)


def vector_to_string(vector: list[float]) -> str:
    '''Serializing the vector as the accepted string format by pgvector.'''

    strings = map(str, vector)
    concatenated = ','.join(strings)

    return f'[{concatenated}]'


def string_to_vector(string: str) -> list[float]:
    '''Deserializing the string as formatted by pgvector back to a list of floats.'''

    concatenated = string.strip('[]')
    strings = concatenated.split(',')
    floats = map(float, strings)

    return list(floats)



from importlib import import_module
from types import ModuleType


def load_module_by_vendor(ai_vendor: str) -> ModuleType:
    '''Dynamically loads the AI module according to the AI vendor string and returns it.'''

    try:
        return import_module(f'sbretrieval.ai_models.{ai_vendor}')

    except ModuleNotFoundError:
        raise ValueError(f'Invalid AI vendor: {ai_vendor}')



class SummaryBasedRetrievalError(Exception):
    '''Encapsulates any error caught by the library.'''
    def __init__(self, message: str) -> None:
        super().__init__(message)


class DocumentIndexingError(SummaryBasedRetrievalError):
    '''Raised when an error occurs during an indexing operation.'''
    def __init__(self) -> None:
        super().__init__('An error happened while indexing a document, check its cause for more details.')


class DocumentRemovalError(SummaryBasedRetrievalError):
    '''Raised when an error occurs during a removal operation.'''
    def __init__(self) -> None:
        super().__init__('An error happened while removing a document, check its cause for more details.')


class DocumentSearchError(SummaryBasedRetrievalError):
    '''Raised when an error occurs during a search operation.'''
    def __init__(self) -> None:
        super().__init__('An error happened while searching documents, check its cause for more details.')



from typing import cast
import os

assert 'OPENAI_API_KEY' in os.environ, 'The environment variable OPENAI_API_KEY is not set'


async def count_tokens(content: str, model: str) -> int:
    '''Counts the tokens of the given content according to the given model.'''

    from tiktoken import encoding_for_model

    encoding = encoding_for_model(model)
    tokens = encoding.encode(content)

    return len(tokens)


async def generate_embedding(content: str, model: str) -> list[float]:
    '''Generates an embedding of the given content using the given model.'''

    from openai import AsyncOpenAI

    client = AsyncOpenAI()
    response = await client.embeddings.create(model=model, input=content, encoding_format='float')
    response_content = response.data[0].embedding

    return response_content


async def send_prompt(prompt: str, model: str) -> str:
    '''Sends a prompt to the given model and returns its response.'''

    from openai import AsyncOpenAI

    messages = [{'role': 'user', 'content': prompt}]

    client = AsyncOpenAI()
    response = await client.chat.completions.create(model=model, messages=messages)  # type: ignore[arg-type]
    response_content = response.choices[0].message.content

    return cast(str, response_content)



import os

assert 'GEMINI_API_KEY' in os.environ, 'The environment variable GEMINI_API_KEY is not set'


async def count_tokens(content: str, model: str) -> int:
    '''Counts the tokens of the given content according to the given model.'''

    from google.generativeai import GenerativeModel  # type: ignore[import-untyped]

    client = GenerativeModel(model)
    response = client.count_tokens(content)

    return response.total_tokens


async def generate_embedding(content: str, model: str) -> list[float]:
    '''Generates an embedding of the given content using the given model.'''

    from google.generativeai import embed_content_async

    arguments = {
        'model': model,
        'content': content
    }

    response = await embed_content_async(**arguments)
    response_content = response['embedding']

    return response_content


async def send_prompt(prompt: str, model: str) -> str:
    '''Sends a prompt to the given model and returns its response.'''

    from google.generativeai import GenerativeModel

    client = GenerativeModel(model)
    response = await client.generate_content_async(prompt)
    response_content = response.text

    return response_content



from typing import Optional

from asyncpg import Connection  # type: ignore[import-untyped]

from sbretrieval.ai_models import load_module_by_vendor
from sbretrieval.database import delete_document, search_documents_by_embeddings, upsert_document, \
    upsert_document_embedding, upsert_document_summary, SearchResult
from sbretrieval.exceptions import DocumentIndexingError, DocumentRemovalError, DocumentSearchError


# default values assumed by the library and assignable by the user
default_ai_vendor = 'openai'
default_embedding_model = 'text-embedding-ada-002'
default_summary_model = 'gpt-4o-mini'
default_summary_prompt = 'Gere um sumário sem nenhuma formatação do texto abaixo:\n\n{}'


async def index_document(
    connection: Connection,
    document_key: str,
    document_content: str,
    ai_vendor: Optional[str] = None,
    embedding_model: Optional[str] = None,
    summary_model: Optional[str] = None,
    summary_prompt: Optional[str] = None
) -> None:
    '''
    Takes a document key and content, count the tokens of the content, generates a summary, generates an
    embedding of the summary, and save both the received and generated data.
    '''

    try:

        # assigning the default values at the module level if none was provided on this call
        ai_vendor = ai_vendor or default_ai_vendor
        embedding_model = embedding_model or default_embedding_model
        summary_model = summary_model or default_summary_model
        summary_prompt = summary_prompt or default_summary_prompt

        # checking the given parameters
        assert connection, 'Connection was not provided.'
        assert document_key, 'Document key was not provided.'
        assert document_content, 'Document content was not provided.'

        # checking even the ones backed by a default (just in case)
        assert ai_vendor, 'AI vendor was not provided (did you unset the default?).'
        assert embedding_model, 'Embedding model was not provided (did you unset the default?).'
        assert summary_model, 'Summary model was not provided (did you unset the default?).'
        assert summary_prompt, 'Summary prompt was not provided (did you unset the default?).'

        # loading the AI module according to the AI vendor string
        ai_module = load_module_by_vendor(ai_vendor)

        # getting the token count for the document content
        document_token_count = await ai_module.count_tokens(document_content, summary_model)

        # requesting a summary of the document to the LLM
        formatted_prompt = summary_prompt.format(document_content)
        document_summary = await ai_module.send_prompt(formatted_prompt, summary_model)

        # generating an embedding for the summary
        document_embedding = await ai_module.generate_embedding(document_summary, embedding_model)

        # upserting everything we generated into the database
        async with connection.transaction():

            await upsert_document(connection, document_key, document_token_count)

            await upsert_document_summary(
                connection, document_key, document_summary, ai_vendor, summary_model
            )

            await upsert_document_embedding(
                connection, document_key, document_embedding, ai_vendor, embedding_model
            )

    except Exception as e:

        # re-raising the error with the library's own exception type
        raise DocumentIndexingError() from e


async def remove_document(connection: Connection, document_key: str) -> bool:
    '''
    Removes from the database all indexed information from the given document.
    Returns true if all the database records were found and deleted and false if any of them were missing.
    '''

    try:

        # checking the given parameters
        assert connection, 'Connection was not provided.'
        assert document_key, 'Document key was not provided.'

        # deletes all records from all tables related to this document key
        async with connection.transaction():
            return await delete_document(connection, document_key)

    except Exception as e:

        # re-raising the error with the library's own exception type
        raise DocumentRemovalError() from e


async def search_documents(
    connection: Connection,
    query: str,
    max_token_count: int,
    ai_vendor: Optional[str] = None,
    embedding_model: Optional[str] = None
) -> SearchResult:
    '''
    Takes a query and generates an embedding of it, then searches for similar documents (by embedding
    proximity).
    '''

    try:

        # assigning the default values at the module level if none was provided on this call
        ai_vendor = ai_vendor or default_ai_vendor
        embedding_model = embedding_model or default_embedding_model

        # checking the given parameters
        assert connection, 'Connection was not provided.'
        assert query, 'Query was not provided.'
        assert max_token_count > 0, 'Max token count should be a positive integer.'

        # checking even the ones backed by a default (just in case)
        assert ai_vendor, 'AI vendor was not provided (did you unset the default?).'
        assert embedding_model, 'Embedding model was not provided (did you unset the default?).'

        # loading the AI module according to the AI vendor string
        ai_module = load_module_by_vendor(ai_vendor)

        # generating an embedding for the summary
        document_embedding = await ai_module.generate_embedding(query, embedding_model)

        # performing the embedding search on the database and returning the results
        async with connection.transaction():
            return await search_documents_by_embeddings(connection, document_embedding, max_token_count)

    except Exception as e:

        # re-raising the error with the library's own exception type
        raise DocumentSearchError() from e
